@@ -13,6 +13,8 @@ import os
 import numpy as np
 import pandas as pd
 import random
+from tqdm import tqdm
+from pqdm.threads import pqdm  
 
 class cognet:
     """Aggregate related Qnet functions
@@ -29,10 +31,10 @@ class cognet:
         self.all_samples = None
         self.samples = None
         self.samples_as_strings = None
-        self.features = None
-        self.cols = None
-        self.immutable_vars = None
-        self.mutable_vars = None
+        self.features = []
+        self.cols = []
+        self.immutable_vars = []
+        self.mutable_vars = []
         self.poles = None
         self.polar_features = None
         self.polar_indices = None
@@ -52,7 +54,8 @@ class cognet:
                         data_obj,
                         key,
                         im_vars=None,
-                        m_vars=None):
+                        m_vars=None,
+                        verbose=False):
         """load parameters from model object
 
         Args:
@@ -61,6 +64,7 @@ class cognet:
           key (str): 'all', 'train', or 'test', corresponding to sample type
           im_vars (list[str], optional): Not implemented yet. Defaults to None.
           m_vars (list[str], optional): Not implemented yet. Defaults to None.
+          verbos (bool, optional): Whether or not to print out model state. Defaults to False. 
         """
         if model is not None:
             # inherit atrributes from model object
@@ -96,6 +100,10 @@ class cognet:
                 variation_weight.append(entropy(v,base=len(v)))
             variation_weight = np.nan_to_num(variation_weight) # remove nans
             self.variation_weight = variation_weight
+        if verbose:
+            print("total features: " + str(len(self.features.columns)) + "\n" 
+                  + "mutable features: " + str(len(self.mutable_vars.columns)) + "\n"
+                  + "immutable features: " + str(len(self.immutable_vars)))
     
     def load_from_dataformatter(self, 
                                 data_obj,
@@ -178,14 +186,29 @@ class cognet:
         self.mutable_vars = [x for x in self.cols
                             if x.upper() not in self.immutable_vars.columns]
     
+    def model_state(self):
+        '''returns state of model
+        '''
+        immutable_var_len = 0
+        if self.immutable_vars:
+            immutabe_var_len =  len(self.immutable_vars.columns)
+        print("total features: " + str(len(self.features.columns)) + "\n" 
+                + "mutable features: " + str(len(self.mutable_vars.columns)) + "\n"
+                + "immutable features: " + str(immutable_var_len))
+        
+        print("total samples: " + str(len(self.samples)))
+        
+        
     def set_nsamples(self,
                     num_samples,
-                    random=False):
+                    random=False,
+                    verbose=True):
         '''select a subset of the samples
 
         Args:
           num_samples (int): Set num of samples to subset, default to None, resets to all samples
-          random (bool): take random sample if true, ordered sample if false
+          random (bool): take random sample if true, ordered sample if false. Defaults to False
+          verbose (bool): whether or not to print out model state regarding samples. Defaults to True.
         '''
         # each time function is called, reset samples to use_all_samples
         # this allows us to call nsamples numerous times 
@@ -213,9 +236,15 @@ class cognet:
                     self.samples = self.samples.iloc[:num_samples]
                 self.nsamples = num_samples
                 self.samples_as_strings = self.samples[self.cols].fillna('').values.astype(str)[:]
+                if verbose:
+                    if random:
+                        print("The number of random samples have been set to " + str(num_samples))
+                    else:
+                        print("The number of samples have been set to " + str(num_samples))
                 
             elif self.samples is None:
                 raise ValueError("load_data first!")
+            
 
     def __variation_weight(self,
                         index):
@@ -271,29 +300,53 @@ class cognet:
             return qsample(sample,self.qnet,steps)
 
     def random_sample(self,
+                      type="prob",
                       df=None,
-                      n=1):
+                      n=1,
+                      steps=200,
+                      n_jobs=3):
         '''compute a random sample from the underlying distributions of the dataset, by column.
         
         
         Args:
+          type (str): How to randomly draw samples. Can take on "null", "uniform", or "prob". Deafults to "prob".
           df (pandas.DataFrame): Desired data to take random sample of. Defaults to None, in which case qnet samples are used.
           n (int): number of random samples to take. Defaults to 1.
+          steps (int): number of steps to qsample. Defaults to 1000
           
         Returns:
-          return_df (pd.DataFrame): Random sample drawn from underlying distribution of each column.
+          return_df (pd.DataFrame): Drawn random sample.
         '''
-        # check if a new dataset was inputted
+        # check if a new dataset was given
         if df is None:
             samples_ = self.samples
         else:
             samples_ = df
 
-        # take random sample from each of the columns based on their distribution
         return_df = pd.DataFrame()
-        for col in samples_.columns:
-            return_df[col] = samples_[col].sample(n=n, replace=True).values
+        # take random sample from each of the columns based on their probability distribution
+        if type == "prob":
+            for col in samples_.columns:
+                return_df[col] = samples_[col].sample(n=n, replace=True).values
+                
+        # random sampling using Qnet qsampling
+        elif type == "null":
+            null_array = np.zeros((len(samples_.columns),), dtype=str)
+            args = [[null_array, steps] for i in range(n)]
+            qsamples = pqdm(args, self.qsampling, n_jobs=n_jobs, argument_type='args') 
             
+            # for i in range(n):
+            #     qsamples.append(self.qsampling(null_array, steps))
+            return_df = pd.DataFrame(qsamples, columns=samples_.columns)
+            
+        # random sampling using uniform distribution of values by Columns
+        elif type == "uniform":
+            for col in samples_.columns:
+                # get unqiue values for each column and draw n values randomly
+                values = samples_[col].unique().astype(str)
+                return_df[col]=np.random.choice(values, size=n, replace=True)
+        else:
+            raise ValueError("Type is not supported!")
         return return_df
     
     def set_poles(self,
@@ -303,7 +356,7 @@ class cognet:
                   steps=0,
                   mutable=False,
                   VERBOSE=False,
-                  restrict=True,
+                  restrict=False,
                   nsamples = None,
                   random=False):
         '''set the poles and samples such that the samples contain features in poles
@@ -315,7 +368,7 @@ class cognet:
           pole_2 (str): column name for second pole
           mutable (bool): Whether or not to set poles as the only mutable_vars
           VERBOSE (bool): boolean flag prints number of pole features not found in sample features if True
-          restrict (bool): boolean flag restricts the sample features to polar features if True
+          restrict (bool): boolean flag restricts the sample features to polar features if True. Defaults to False.
           random (bool): boolean flag takes random sample of all_samples
         '''
         invalid_count = 0
@@ -488,25 +541,41 @@ class cognet:
     
     def distfunc_multiples(self,
                            outfile,
-                           processes=6):
+                           processes=6,
+                           samples=None):
         """compute distance matrix for all samples in the dataset
 
         Args:
           outfile (str): desired output filename and path
+          processes (int): Number of processes to run in parallel. Defaults to 6.
+          samples (2D array): Dataset from which to calculate qdist matrix. Defaults to None.
           
         Returns:
           result: pandas.DataFrame containing distance matrix
         """
         if all(x is not None for x in [self.samples, self.features]):
+            # if exterior dataset is given, temporarily replace class attributes
+            if samples is not None:
+                original_samples = self.samples
+                original_samples_as_strings = self.samples_as_strings
+                self.samples = samples
+                samples = samples.fillna("").values.astype(str)
+                self.samples_as_strings = samples
             cols = [i for i in range(len(self.samples))]
             result = self.mp_compute(processes,
                                         self.distfunc_line,
                                         cols,
                                         outfile)
+            
             # format and save resulting dict, and tranpose symmetrical distance matrix
             result = result.to_numpy()
             result = pd.DataFrame(np.maximum(result, result.transpose()))
             result.to_csv(outfile, index=None, header=None)
+            
+            # replace class attributes with originals
+            if samples is not None:
+                self.samples = original_samples
+                self.samples_as_strings = original_samples_as_strings
         else:
             raise ValueError("load data first!")
         
@@ -599,7 +668,9 @@ class cognet:
         '''
         # if all(x is not None for x in [self.year]):
             # init file names 
-        yr = self.year
+        yr = ''
+        if self.year is not None:
+            yr = self.year
         PREF = name_pref
         FILE = infile
         DATAFILE = out_dir + 'data_' +yr
@@ -1064,12 +1135,15 @@ class cognet:
         args=[None, index_colname, output_dir,
               file_name, mask_prob, allow_all_mutable, 
               save_samples, save_output]
+
         
         result = self.mp_compute(processes,
-                                    self.randomMaskReconstruction,
-                                    cols,
-                                    outfile,
-                                    args=args)
+                                 self.randomMaskReconstruction,
+                                 cols,
+                                 outfile,
+                                 args=args)
+
+        
         return result
     
     def dmat_filewriter(self,
@@ -1188,7 +1262,7 @@ class cognet:
                                "# time requested\n",
                                "T={}\n".format(T),
                                "NUM=\'all\'\n",
-                               "LAUNCH=\'{}\'\n\n".format(MPI_LAUNCHER_FILE),
+                               "LAUNCH=./\'{}\'\n\n".format(MPI_LAUNCHER_FILE),
                                "for yr in `echo $YEARS`\n",
                                "do\n",
                                "\techo $yr\n",
@@ -1199,4 +1273,4 @@ class cognet:
             os.system("cp {} {}".format(MPI_LAUNCHER_FILE,tmp_path+'mpi_launcher.sh'))
         
         else:
-            raise ValueError("load data first!")
\ No newline at end of file
+            raise ValueError("load data first!")
